##############################################################################
#
# License
# -------
#
# Copyright (c) Gerard Hale. All right reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#
# 3. Gerard Hale requests that attribution be given to Zope
#    in any manner possible. A significant investment has been put
#    into Zope, and this effort will continue if the Zope community
#    continues to grow. This is one way to assure that growth.
#
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
#
#      "This product includes software developed by Gerard Hale
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
#
# 5. Names associated with Gerard Hale must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Gerard Hale.
#
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
#
#      "This product includes software developed by Gerard Hale
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
#
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official software releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.
#
#
# Disclaimer
#
#   THIS SOFTWARE IS PROVIDED BY Gerard Hale ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL Gerard Hale OR HIS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
#
#
# This software consists of contributions made by Gerard Hale and
# many other individuals.
#
##############################################################################
"""Z Session product """
__version__='$Revision: 0.0.8'

import Globals
from Globals import Persistent
from Globals import HTMLFile, MessageDialog
from string import join,strip,split,lower
import Acquisition
from Acquisition import Implicit, Explicit
import OFS.SimpleItem
from OFS.Document import Document
from urllib import quote, unquote
import AccessControl.Role
import string
from cPickle import loads, dumps
import time
import threading
import os

#The global Session Dictionary.
SessionDict={}

class ZSessionObj(Explicit):
    """ A ZSession Object. Installed into the REQUEST object as 
    REQUEST.SESSION (like REQUEST.RESPONSE)
    """

    __allow_access_to_unprotected_subobjects__ = 1

    ZSession = Acquisition.Acquired

    def __init__(self, parentSession, sessionName, mode='base64'):
        self.sName = sessionName
        if SessionDict.has_key(sessionName):
            self.cachedict = SessionDict[sessionName]
        else:
            self.cachedict = {}
        self.cachedict[self.dumper('TimeStamp')] = self.dumper(int(time.time()))

    def getName(self):
        "return name of session"
        return self.sName

    def loader(self, v):
        "load - base64"
        return loads(decodestring(v))

    def dumper(self, v):
        "dump - base64"
        return encodestring(dumps(v))[:-1] # chop \n

    def __getitem__(self, name):
        if self.cachedict.has_key(self.dumper(name)):
            try:
                return self.loader(self.cachedict[self.dumper(name)])
            except:
                return self.cachedict[self.dumper(name)]
        else:
            return ""

    def __setitem__(self, name, value):
        #print "setitem called for %s of type %s"%(name, type(value))
        #if SESSION value is an object do not encrypt it.
        try:
            self.cachedict[self.dumper(name)] = self.dumper(value)
        except:
            self.cachedict[self.dumper(name)] = value
        SessionDict[self.getName()] = self.cachedict
        return None

    set = __setitem__

    def __delitem__(self, name):
        del self.cachedict[self.dumper(name)]
        SessionDict[self.getName()] = self.cachedict

    delete = __delitem__

    def __len__(self):
        return len(self.cachedict)

    def keys(self):
        l = []
        for r in self.cachedict.keys():
           l.append(self.loader(r))
        return l

    def values(self):
        l = []
        for v in self.cachedict.values():
            l.append(self.loader(v))
        return l

    def items(self):
        return self.cachedict.items()

    def has_key(self, name):
        if self.cachedict.has_key(self.dumper(name)):
            return 1
        else:
            return 0

    def __str__(self):
        l = ["ZSession Variables:\n"]
        for n,v in self.items():
            l.append("%s: %s"%(n,str(v)))
        return join(l, '\n')

    def clear(self):
        self.cachedict.clear()

    def copy(self):
        raise RuntimeError, "copy() unimplemented for ZSessions"

    def update(self, dict):
        for k, v in dict.items():
            self[k] = v

    def get(self, key, failobj=None):
        if self.has_key(key): return self[key]
        else: return failobj    


class ZSession(OFS.SimpleItem.Item,
                 Persistent,
                 Implicit,
                 AccessControl.Role.RoleManager
                 ):
    """ A ZSession. """

    meta_type = 'ZSession'
    id        = 'ZSession'
    title     = 'ZSession'
    icon      = 'misc_/ZSession/ZSession_icon'

    manage_options=(
    {'label':'Sessions',     'action':'manage_sessions'},
    {'label':'Properties',   'action':'manage_main'},
    {'label':'Process',      'action':'manage_process'},
    {'label':'Security',     'action':'manage_access'},
    )

    __ac_permissions__=(
        ('View management screens',
         ['manage','manage_menu','manage_main','manage_tabs' ]),
        ('Change ZSessions',     ['manage_edit']),
        ('Clear ZSessions',     ['manage_sessions']),
        ('View', ['__call__','']),
    )

    manage_main=HTMLFile('dtml/main', globals())
    manage_sessions=HTMLFile('dtml/manage_sessions', globals())
    manage_session_details=HTMLFile('dtml/manage_session_details', globals())
    manage_process=HTMLFile('dtml/manage_process', globals())

    def __init__(self, cookie_name, SessionTimeout, mode=None, path='/'):
        self.cookie_name = cookie_name
        self.SessionTimeout = SessionTimeout
        if not mode: mode = 'base64'
        self.mode = mode
        self.path = path
        self.lastRunTime = time.asctime(time.localtime(time.time()))
        self.sessionsCleared = []

    def manage_edit(self, cookie_name, mode, path, SessionTimeout, REQUEST=None):
        """Change properties"""
        self.cookie_name = cookie_name
        self.SessionTimeout = SessionTimeout
        self.mode = mode
        self.path = path
        ZSession.lastRunTime = ZSession.lastRunTime
        ZSession.sessionsCleared = ZSession.sessionsCleared
        if REQUEST:
            return MessageDialog(
                title   ='ZSession Changed',
                message ='ZSession properties have been updated',
                action  = REQUEST['URL1']+'/manage_main',
                target  ='manage_main')

    def __call__(self, REQUEST=None, session_id=None, noCookie=None, validSession=None):
        """ Method called checks or sets cookie, creates new ZSession Object. """
        if REQUEST is None:
            try:
                REQUEST=self.REQUEST
            except:
                raise "Session Error", "can't find REQUEST, even in self!"
        response = REQUEST['RESPONSE']

        if session_id is not None:
            if not REQUEST.cookies.has_key(self.cookie_name):
                sessionName = self.newSession(REQUEST, session_id, noCookie)
            sessionName = REQUEST.cookies[self.cookie_name]
            if sessionName != session_id:
                sessionName = self.newSession(REQUEST, session_id, noCookie)
        else:
            if not REQUEST.cookies.has_key(self.cookie_name):
                sessionName = self.newSession(REQUEST, noCookie=noCookie)
            else:
                if ( REQUEST.has_key('_force_new_session') and
                     REQUEST['_force_new_session'] ):
                    sessionName = self.newSession(REQUEST, noCookie=noCookie)
                else:
                    sessionName = REQUEST.cookies[self.cookie_name]
        self.buildSessionObj(REQUEST, sessionName)
        blnThreadFound = 0
        counter = 0
        for t in threading.enumerate():
            if (t.getName() == "ZSession"):
                blnThreadFound = 1
                counter = counter + 1
        if(blnThreadFound==0):
            logProcess("__call__: Clear Sessions Thread has not started or exited!")
            cleanup=threading.Thread(target=self.clearSessionsProcess, name='ZSession', args=())
            cleanup.setDaemon(1) # so thread will terminate with parent process
            cleanup.start()
        #If dtml-var is used print Session Name
        return sessionName

    def buildSessionObj(self, REQUEST, sessionName):
        s = ZSessionObj(self, sessionName, self.mode)
        REQUEST['SESSION'] = s.__of__(self)

    def newSession(self, REQUEST, sessionName=None, noCookie=None):
        " create a new session and set cookie. "
        response = REQUEST['RESPONSE']
        if sessionName is None:
            sessionName = self.makeSessionName(REQUEST)
            SessionDict[sessionName]={}
        if not noCookie:
            response.setCookie(self.cookie_name, sessionName, path=self.path)
        REQUEST.cookies[self.cookie_name] = sessionName
        self.buildSessionObj(REQUEST, sessionName)
        return sessionName

    def makeSessionName(self, REQUEST):
        """ create a new unique string to name the session. feel free to
            fiddle with this. """
        import md5, time, base64
        m = md5.new()
        m.update('A unique string is added here???')
        m.update(str(time.time()))
        m.update(str(REQUEST))
        return string.replace(base64.encodestring(m.digest())[:-3], '/', '$')

    def getSessionData(self, SessionName=None):
        retDict = {}
        tmpDict = {}
        tmpSessionList = []
        if SessionName:
            tmpDict = SessionDict[SessionName]
            for n,v in tmpDict.items():
                try:
                    retDict[self.loader(n)] = self.loader(v)
                except:
                    retDict[n] = v
            sd = SessionName, retDict
            tmpSessionList.append(sd)
        else:
            for k in SessionDict.keys():
                retDict = {}
                tmpDict = SessionDict[k]
                for n,v in tmpDict.items():
                    try:
                        retDict[self.loader(n)] = self.loader(v)
                    except:
                        retDict[n] = v
                sd = k, retDict
                tmpSessionList.append(sd)
        return tmpSessionList

    def showAllSessions(self):
        SessionList = []
        if(SessionDict=={}):
            return []
        else:
            for k in SessionDict.keys():
                dict = SessionDict[k]
                if dict.has_key(self.dumper('TimeStamp')):
                    ts = self.loader(dict[self.dumper('TimeStamp')])
                    SessionList.append([k,time.asctime(time.localtime(ts))])
        return SessionList

    def manage_clearSessions(self, SessionNames=[], REQUEST=None):
        """Clear Session Data from Session Dictionary
        The Session specified in 'SessionNames' gets deleted.
        """
        if type(SessionNames) is type(''): SessionNames=[SessionNames]
        if not SessionNames:
            return MessageDialog(title='No items specified',
                   message='No items were specified!',
                   action ='./manage_sessions',)
        for SessionName in SessionNames:
            if SessionDict.has_key(SessionName):
                del SessionDict[SessionName]
        if REQUEST is not None:
            return self.manage_sessions(self, REQUEST)

    def clearSessionsProcess(self):
        """ clears old sessions that have timed out """
        sleepSeconds = 15
        while 1:
            sessionsClearedList = []
            if(SessionDict!={}):
                for k in SessionDict.keys():
                    dict = SessionDict[k]
                    if dict.has_key(self.dumper('TimeStamp')):
                        SessionTime = self.loader(dict[self.dumper('TimeStamp')]) + int(self.SessionTimeout)*60
                        if (SessionTime<int(time.time())):
                            del SessionDict[k]
                            sessionsClearedList.append(k)
            ZSession.lastRunTime = time.asctime(time.localtime(time.time()))
            ZSession.sessionsCleared = sessionsClearedList
            time.sleep(sleepSeconds)

    def showProcessLog(self):
        """
        Only one thread should ever be running, but if for some reason there are more
        they will be added to the list and show in the manage process screen.
        """
        threadList = []
        for t in threading.enumerate():
            if (t.getName() == "ZSession"):
                threadProperties = {}
                threadProperties['getName'] = t.getName()
                threadProperties['isAlive'] = t.isAlive()
                threadProperties['isDaemon'] = t.isDaemon()
                """This assumes there is only one thread."""
                try:
                    threadProperties['lastRunTime'] = ZSession.lastRunTime
                    threadProperties['sessionsCleared'] = ZSession.sessionsCleared
                except:
                    threadProperties['lastRunTime'] = 'Error: ZSession not updated yet'
                    threadProperties['sessionsCleared'] = 'Error: ZSession not updated yet'
                threadList.append(threadProperties)
        return threadList

    def loader(self, v):
        "load - base64"
        return loads(decodestring(v))

    def dumper(self, v):
        "dump - base64"
        return encodestring(dumps(v))[:-1] # chop \n


#Global Functions

def logProcess(msg):
    LogFile = os.path.join(Globals.data_dir, 'ZSession.log')
    if(os.path.exists(LogFile)):
        f = open(LogFile, 'a')
        f.write(time.asctime(time.localtime(time.time())) + ' : ' + msg + '\n')
        f.close()
    else:
        """File does not exist yet"""
        try:
            f = open(LogFile, 'w')
            f.write(time.asctime(time.localtime(time.time())) + ' : ' + msg + '\n')
            f.close()
        except:
            #can not write to var directory??
            pass

import binascii, base64, string

def encodestring(bin):
    b2a = binascii.b2a_base64
    MAXLINESIZE = 76 # Excluding the CRLF
    MAXBINSIZE = (MAXLINESIZE/4)*3
    if len(bin) > MAXBINSIZE:
        out = []
        for o in range(0, len(bin), MAXBINSIZE):
            out.append(b2a(bin[o:o+MAXBINSIZE]))
        asc = string.join(out,'')
    else:
        asc = b2a(bin)
    return asc

def decodestring(asc):
    a2b = binascii.a2b_base64
    bin = a2b(asc)
    return bin
